%%%%
% Machine learning system for experimental purposes
% PLMS = Prolog Machine Learning System
%
% Author: Tony Lindgren 
%
% version 0.0.7
%
%  

:- module(pmls,[run_experiment/3, start_experiment/0]).

:- use_module([library(csv), library(lists), library(random), library(codesio),
	       write_results, util, mat, read_from_file, handle_rules,
	       post_process_rules, do_induction, indexing, sampling_methods]).

%%increase memory
:- set_prolog_stack(global, limit(9 000 000 000 000 000 000)).
:- set_prolog_stack(trail,  limit(9 000 000 000 000 000 000)).
:- set_prolog_stack(local,  limit(9 000 000 000 000 000 000)).

start_experiment:-
    experiment(E),
    method(M),
    data(D),
    write('Starting experiment'),nl,
    start_mt,
    run_experiment(experiment(E), method(M), data(D)),
    close_mt.

start_mt:-
    close_mt,     %if we are running	
    current_prolog_flag(cpu_count, NoCores),
    NoCoresForPool is NoCores - 2,
    write('Maximum threads used simultanously is : '),write(NoCoresForPool),nl,
    message_queue_create(ensQueue),
    thread_pool_create(ensPool, NoCoresForPool, []).
      
close_mt:-
   (current_thread_pool(ensPool)->
      message_queue_destroy(ensQueue),
      thread_pool_destroy(ensPool)
   ;
      true
   ).

run_experiment(experiment(Parameters), method(Methods), data(DataSets)):-
   (member(seed(Seed), Parameters)->
    set_random(seed(Seed))                            %use it if we got it!
   ;
    true),
    process_data(DataSets, Parameters, Methods).

process_data([], _, _).
process_data([set(Name, prolog, InfoList)|Sets], Parameters, Methods):-
    write('Data set: '),write(Name),nl,	
    interpret_info(InfoList, Head, BackgroundKnowledge, Examples),	
    read_from_file(BackgroundKnowledge, BKList),
    read_from_file(Examples, ExList),
    random_permutation(ExList, RndExL),	                        %re-order examples
    member(fold_x_v(Folds), Parameters),
    pre_process_data(Folds, Parameters, RndExL, ExLists),
    (member(cost_matrix(CM), InfoList) ->
      use_methods(Methods, Folds, Name, Head, [cost_matrix(CM)|Parameters], no_f, ExLists, BKList)
    ;
      use_methods(Methods, Folds, Name, Head, Parameters, no_f, ExLists, BKList)
     ),!,
    process_data(Sets, Parameters, Methods).
process_data([set(Name, csv, CsvParameters)|Sets], Parameters, Methods):-
    write('Data set: '),write(Name),nl,
    get_csv_examples(CsvParameters, Parameters, NewParameters, Folds, Head, ExLists, FeaturesDTL),
    (member(cost_matrix(CM), CsvParameters) ->
      use_methods(Methods, Folds, Name, Head, [cost_matrix(CM)|NewParameters], FeaturesDTL, ExLists, no_bk)
    ;
      use_methods(Methods, Folds, Name, Head, NewParameters, FeaturesDTL, ExLists, no_bk)
     ),!,
    process_data(Sets, Parameters, Methods).

use_methods([], _, _, _, _, _, _, _).
use_methods([Method|Methods], Folds, Name, Head, Parameters, FeaturesDTL, ExLists, BKList):-
    %trace,	
    do_experiment(Folds, Name, Head, Method, ExLists, FeaturesDTL, BKList, Results),
    open('result', append, Stream),
    %write('before write results'),nl,%trace,
    write_results(Results, Stream, Parameters, Folds),
    close(Stream),
    %write('after write results'),nl,
    retract_loop(Results, Head, BKList),!,
    use_methods(Methods, Folds, Name, Head, Parameters, FeaturesDTL, ExLists, BKList).

retract_loop([], _, _).
retract_loop([result(_, _, _, Key, _, _, _, _, _)|Keys], Head, BKList):-
    retract_inner_loop(Key, Head, BKList),
    retract_loop(Keys, Head, BKList).

retract_inner_loop(Key, Head, BKList):-
    atom(Key),
    retractall(Key:Head),
    retract_all(BKList, Key).
retract_inner_loop([], _, _).
retract_inner_loop([Key|Keys], Head, BKList):-
    retractall(Key:Head),
    retract_all(BKList, Key),
    retract_inner_loop(Keys, Head, BKList).
retract_inner_loop(_, _, _). %Fail silently

do_experiment(0, _, _, _, _, _, _, []).
do_experiment(Folds, Name, Head, Parameters, ExLists, FeaturesDTL, BKList, Eval):-
    write('Do fold: '),write(Folds),nl,	
    do_one_fold(Folds, Name, Head, Parameters, ExLists, FeaturesDTL, BKList, PartEval),
    NewFolds is Folds - 1,!,
    do_experiment(NewFolds, Name, Head, Parameters, ExLists, FeaturesDTL, BKList, RestEval),
    append(PartEval, RestEval, Eval).
do_experiment(Folds, Name, Head, Parameters, ExLists, FeaturesDTL, BKList, Eval):-
    write(Folds), write(Name), write(Head), write(Parameters), %write(ExLists), write(BKList), write(Eval),
    do_experiment(Folds, Name, Head, Parameters, ExLists, FeaturesDTL, BKList, Eval).

%CSV - classification
do_one_fold(TestFold, Name, Head, Method, ExLists, FeaturesDTL, no_bk, Result):-
    TestFold \== 0,
    nth1(TestFold, ExLists, TestExList, TrainExList),
    do_induction(Method, Head, TrainExList, no_bk, TestFold, ClassIndex, FeaturesDTL, Keys, _RawInduction, NoRules1, ClassT, Rules),
    %write('clean_con'),nl,
    clean_conditions(Rules, CleanCondRulesL),
    %write('add_rule_id'),nl,
    add_rule_id(CleanCondRulesL, RulesWId, NoConds, NoRules1, NoRules),
    %write('post_process_rules'),nl,
    post_process_rules(Name, Head, TestFold, Method, TestExList, TrainExList, no_bk, ClassIndex, Keys, NoRules, NoConds, ClassT, RulesWId, Result).

do_one_fold(TestFold, Name, Head, Method, ExLists, no_f, BKList, Result):-
    TestFold \== 0,
    nth1(TestFold, ExLists, TestExList, TrainExList),
    do_induction(Method, Head, TrainExList, BKList, TestFold, ClassIndex, no_f, Keys, _RawInduction, NoRules1, ClassT, Rules),!,    
    clean_conditions(Rules, CleanCondRulesL),
    add_rule_id(CleanCondRulesL, RulesWId, NoConds, NoRules1, NoRules),
    post_process_rules(Name, Head, TestFold, Method, TestExList, TrainExList, BKList, ClassIndex, Keys, NoRules, NoConds, ClassT, RulesWId, Result).

%Dead code?
/*
evaluate_tree(Type, TestExList, DataName, SampMethod, SampleSize, Tree, BKList, Head, TreeKey, ClassIndex, ClassT,
	      result(DataName, SampMethod, SampleSize, TreeKey, ClassT, NoRules, TreeEval, TreeEvalTime)):-
    assert_tree_to_bb(Tree, TreeKey, NoRules, _TreeRules),
    statistics(runtime, _),
    classify_test_set(Type, TestExList, ClassIndex, ClassT, ClassT, ClassDist, [TreeKey], TreeRes),
    statistics(runtime, [_,TreeEvalTime]),
    retract_inner_loop(TreeKey, Head, BKList),
    length(TreeRes, TreeResL),
    copy_term(ClassT, CClassT),
    evaluate(ClassDist, CClassT, TreeRes, TreeResL, TreeEval),!.

%remove this?! Dead code...
remove_all_maj_class_rules([], _, _, _, []).
remove_all_maj_class_rules([rule(_,Conds,_,ClassT)|Rest], true, _, MajClass, NewRevTransRules):-
     get_majority_class(ClassT, MajClass),!,
     remove_all_maj_class_rules(Rest, false, [Conds], MajClass, NewRevTransRules).
remove_all_maj_class_rules([rule(_,Conds,_,ClassT)|Rest], false, PrevNConds, MajClass, NewRevTransRules):-
     get_majority_class(ClassT, MajClass),!,
     remove_all_maj_class_rules(Rest, false, [Conds|PrevNConds], MajClass, NewRevTransRules).
remove_all_maj_class_rules([rule(H,C,NC,CT)|Rest], _, PrevNC, MajClass, [rule(H,C,NNC,CT)|NewRevTransRules]):-
     append(PrevNC,NC,NNC),
     remove_all_maj_class_rules(Rest, true, PrevNC, MajClass, NewRevTransRules).

check_ties(Res, NoTies):-
    check_ties(Res, 0, NoTies).
check_ties([], NoTies, NoTies).
check_ties([_-[No-_,No-_|_]|Rest], TempNoT, NoTies):-
    NewTempNoT is TempNoT + 1,!,
    check_ties(Rest, NewTempNoT, NoTies).
check_ties([_|Rest], TempNoT, NoTies):-
    check_ties(Rest, TempNoT, NoTies).
*/

get_forrest_size(ensemble_model(Size, _, _), Size).
get_learning_type(ensemble_model(_, _, Type), Type).

create_rule_id([], _, []).
create_rule_id([Tree|Trees], ID, FinalRulesWithIDs):-
    create_rule_id1(Tree, ID, NewID, RulesWithIDs1),
    create_rule_id(Trees, NewID, RulesWithIDs2),
    append(RulesWithIDs1, RulesWithIDs2, FinalRulesWithIDs).

create_rule_id1([], NewID, NewID, []).
create_rule_id1([Rule|Rules], ID, FinalID, [ID-Rule|RuleWithIDs]):-
    NewID is ID + 1,
    create_rule_id1(Rules, NewID, FinalID, RuleWithIDs).

evaluate([], _, [], _, []).
evaluate([No-Class|CTuples], ClassTuple, Res, ResL, [Class-[No, TrueP, FalseP, FalseN, FalseTup]|Results]):-
    evaluate_class(Res, Class, RestL, 0, 0, 0, ClassTuple, TrueP, FalseP, FalseN, FalseTup),
    %PercentTrueP is (TrueP / No) * 100,   %ResL
    %PercentFalseP is (FalseP / ResL) * 100, %ResL
    %ClassCov is (No / ResL) * 100,
    %nl,
    %write('Class: '), write(Class),nl,
    %write('Covers '), write(ClassCov), write(' of the example space'),nl,
    %write('Accuracy: '), write(PercentTrueP),nl,
    %write('% of False Classifications(wrt. to all classes): '), write(PercentFalseP),nl,
    %write_false_neg1(FalseTup, ResL, FNegL),!,
    evaluate(CTuples, ClassTuple, RestL, ResL, Results).

write_false_neg1(FalseNTup, TotClassified, Tot-FNegL):-
    sum_false(FalseNTup, 0, Tot),
    %Cov is (Tot / TotClassified) * 100,
    %write('% wrongly classified not belonging to the particular class classified as such are: '), write(Cov),nl,
    write_false_neg(FalseNTup, TotClassified, FNegL).

write_false_neg([], _, []).
write_false_neg(_, 0, 0). %:-
    %write('No False negatives classifications'),nl,!.
write_false_neg([No-Class|Rest], Tot, [Class-Percent|FNegL]):-
    Percent is No/Tot * 100,
    %write('Class '), write(Class), write(' wrongly classified in (%, wrt. all classes): '), write(Percent),nl,
    write_false_neg(Rest, Tot, FNegL).

sum_false([], Tot, Tot).
sum_false([No-_|Rest], TTot, Tot):-
    NewTTot is TTot + No,
    sum_false(Rest, NewTTot, Tot).

evaluate_class([], _Class, [], TrueP, FalseP, FalseN, False, TrueP, FalseP, FalseN, False).
evaluate_class([Class-[_-Class|_]|Rest], Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    NewTTrueP is TTrueP + 1,!,
    evaluate_class(Rest, Class, RestL, NewTTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class2-[No-Class|R]|Rest], Class, [Class2-[No-Class|R]|RestL], TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    Class \== Class2,!,
    NewTFalseP is TFalseP + 1,
    update_false(TFalse, Class, _NewTFalse),
    evaluate_class(Rest, Class, RestL, TTrueP, NewTFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class-[_-Class2|_]|Rest], Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    Class \== Class2,!,
    NewTFalseN is TFalseN + 1,
    update_false(TFalse, Class2, NewTFalse),
    evaluate_class(Rest, Class, RestL, TTrueP, TFalseP, NewTFalseN, NewTFalse, TrueP, FalseP, FalseN, False).
evaluate_class([Class2-[No-Class3|R]|Rest], Class, [Class2-[No-Class3|R]|RestL], TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False):-
    evaluate_class(Rest, Class, RestL, TTrueP, TFalseP, TFalseN, TFalse, TrueP, FalseP, FalseN, False).

update_false(TFalseN, Class, NewTFalseN):-
    select(No-Class, TFalseN, RestTFalseN),
    NewNo is No + 1,!,
    append([NewNo-Class], RestTFalseN, NewTFalseN).

%prune - write later if needed
classify_test_set(_Type, [],_ClassIndex, _ClassTuples, FinalCTup, FinalCTup, _Keys, []).
classify_test_set(Type, [TestEx|TestExL], ClassIndex, ClassTuples, TempCTup, FCTup, Keys, [PartRes|Res]):-
    TestEx =..[F|Args],
    nth1(ClassIndex, Args, Class, Rest),
    nth1(ClassIndex, NewArgs, _, Rest),
    %if we have not seen it before add it to tuple, fix this for rest...
    (select(No-Class, TempCTup, TCTup) ->
      NewNo is No + 1,
      append([NewNo-Class], TCTup, NewTempCTup)
    ;
      append([1-Class], TempCTup, NewTempCTup)
    ),
    NTestEx =..[F|NewArgs],
    get_votes(Type, Keys, NTestEx, ClassIndex, Class-ClassTuples, PartRes),!,
    classify_test_set(Type, TestExL, ClassIndex, ClassTuples, NewTempCTup, FCTup, Keys, Res).

get_votes(dac, [], _TestEx, _ClassIndex, TClass-ClassTuples, TClass-RSClassT):-
    sort(ClassTuples, SClassT),
    reverse(SClassT, RSClassT).
get_votes(dac, [Key|Keys], TestEx, ClassIndex, TClass-ClassTuples, CTuples):-
    copy_term(TestEx,CTestEx),
    call(Key:CTestEx),                              %findall fr sac!
    CTestEx =.. [_|Args],
    nth1(ClassIndex, Args, ClassDist),
    keysort(ClassDist,Classes),
    reverse(Classes,[_-Class|_]),
    select(No-Class, ClassTuples, Rest),
    NewNo is No + 1,
    append([NewNo-Class], Rest, NewClassTuples),!,
    get_votes(dac, Keys, TestEx, ClassIndex, TClass-NewClassTuples, CTuples).



loop_write(_,[]).
loop_write(N,[_F|R]):-
    %length(F,L),
    %write('Fold '),write(N), write(' has '), write(L),write(' examples'),nl,
    NewN is N + 1,
    loop_write(NewN,R).

%Could/should be done in conjuntion w induction...
add_rule_id([tree(Tree)|Trees],[tree(Tree)|Trees], _NoConds, NoRules, NoRules).
add_rule_id(RuleSets, RulesWId, NoConds, _, NoRules):-
    add_rule_id1(RuleSets, 0, RulesWId, NoConds, NoRules).

add_rule_id1([], _, [], 0, 0).
add_rule_id1([rule_set(Rules)|RuleSets], RuleSetId, [rule_set(PartRulesWId)|RestRulesWId], FConds, FNoRules):-
    NRuleSetId is RuleSetId + 1,
    add_r_id(Rules, NRuleSetId, 1, PartRulesWId, PartConds, PartNoRules),	   
    add_rule_id1(RuleSets, NRuleSetId, RestRulesWId, RestConds, RestNoRules),
    FConds is PartConds + RestConds,
    FNoRules is PartNoRules + RestNoRules.
add_r_id([], _, _, [], 0, 0).
add_r_id([rule(Cond, Dist)|Rules], NRuleSetId, RuleId, [rule(NRuleSetId-RuleId, Cond, Dist)|RulesWId], FAccConds, FNoRules):-
    NewRuleId is RuleId + 1,
    length(Cond, NoConds),
    add_r_id(Rules, NRuleSetId, NewRuleId, RulesWId, RestAccConds, RestNoRules),
    FAccConds is NoConds + RestAccConds,
    FNoRules is RestNoRules + 1.

add_ex_id(Ex, ExWId):-
   add_e_id(Ex, 1, ExWId).
add_e_id([], _, []).
add_e_id([Ex|Exs], Id, [(Id, Ex)|ExWId]):-
   NewId is Id + 1,	
   add_e_id(Exs, NewId, ExWId).



/*
%%
% fill the predTuple list with values, dead code?!
read_csv_file_to_list(FileName, Name, CsvParameters, ExList):-
  open(FileName, read, Stream),
  %trace,
  handle_input(Stream, 1, Name, CsvParameters, ExList),
  close(Stream).
  
handle_input(Stream, Line, Name, CsvParameters, PredTuple):-
  member(ignore_row(RowList), CsvParameters),
  member(Line, RowList),!,
  read_line_to_codes(Stream, LineOfCodes),
  (LineOfCodes == end_of_file ->
    !
  ;
    NewLine is Line + 1,
    handle_input(Stream, NewLine, Name, CsvParameters, PredTuple)
  ).
handle_input(Stream, Line, Name, CsvParameters, PredTuple):-
  read_line_to_codes(Stream, LineOfCodes),
  (LineOfCodes == end_of_file ->
    !
  ;
    NewLine is Line + 1,
    member(ignore_col(IgCols), CsvParameters),
    member(x_cols(XCols), CsvParameters),
    member(y_cols(YCols), CsvParameters),
    length(XCols, NoXs),
    length(YCols, NoYs),
    Args is NoXs + NoYs,
    functor(PredTuple, Name, Args), 
    handle_line(LineOfCodes, 1, IgCols, XCols, YCols, PredTuple),
    handle_input(Stream, NewLine, Name, CsvParameters, PredTuple)
  ).

%End of the line...
handle_line([], _, _, _, _, _).
%Ignore this column
handle_line(LineOfCodes, Col, IgCol, XCol, YCol, PredTuple):-
  member(Col, IgCol),!,
  get_col(LineOfCodes, _, RestOfCodes),
  NewCol is Col + 1,
  handle_line(RestOfCodes, NewCol, IgCol, XCol, YCol, PredTuple).
%Add this column as x (input)
handle_line(LineOfCodes, Col, IgCol, XCol, YCol, PredTuple):-
  member(Col, XCol),!,
  length(IgCol, RemovedCols),
  get_col(LineOfCodes, ColumnValue, RestOfCodes),
  Arg is Col - RemovedCols,  
  arg(Arg, PredTuple, ColumnValue),
  NewCol is Col + 1,
  handle_line(RestOfCodes, NewCol, IgCol, XCol, YCol, PredTuple).
%Add this column as y (output)
handle_line(LineOfCodes, Col, IgCol, XCol, YCol, PredTuple):-
  member(Col, YCol),!,
  length(IgCol, RemovedCols),
  get_col(LineOfCodes, ColumnValue, RestOfCodes),
  Arg is Col - RemovedCols,  
  arg(Arg, PredTuple, ColumnValue),
  NewCol is Col + 1,
  handle_line(RestOfCodes, NewCol, IgCol, XCol, YCol, PredTuple). 
  

get_col(Codes, Value, RestOfCodes):-
  get_number(Codes, RestOfCodes, Value).
get_col(Codes, Value, RestOfCodes):-
  get_atom(Codes, RestOfCodes, Value).	

%first try to get into a number
get_number(Codes, RestOfCodes, Value):-
  get_numberC(Codes, NumberCodes, RestOfCodes),
  number_codes(Value, NumberCodes).

get_numberC([Code|RestOfCodes], [], RestOfCodes):-
  char_code(',', Code),!.	%Stop at ','
get_numberC([Code|RestOfCodes], NumberCodes, RestOfCodes):-
  char_code(' ', Code),!,	%remove if we have blank junk
  get_numberC(RestOfCodes, NumberCodes, RestOfCodes).
get_numberC([Code|Codes], [Code|NumberCodes], RestOfCodes):-
  get_numberC(Codes, NumberCodes, RestOfCodes).

%else get it as an atom
get_atom(Codes, RestOfCodes, Value):-
  get_atomC(Codes, AtomCodes, RestOfCodes),
  atom_codes(Value, AtomCodes).

get_atomC([Code|RestOfCodes], [], RestOfCodes):-
  char_code(',', Code).
get_atomC([Code|Codes], [Code|AtomCodes], RestOfCodes):-
  get_atomC(Codes, AtomCodes, RestOfCodes).



*/
















